"""GCP Cloud Workflows integration for Django."""

from .exceptions import (
    InvalidWorkflowArgumentsError,
    WorkflowConfigurationError,
    WorkflowError,
    WorkflowExecutionError,
    WorkflowNotFoundError,
)
from .workflows import Workflow, WorkflowExecution

__all__ = [
    "Workflow",
    "WorkflowExecution",
    "WorkflowError",
    "WorkflowNotFoundError",
    "WorkflowExecutionError",
    "InvalidWorkflowArgumentsError",
    "WorkflowConfigurationError",
]



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

from datetime import datetime, timedelta
import gzip
import mimetypes
from unittest import mock
from zoneinfo import ZoneInfo

from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.test import TestCase, override_settings
from django.utils import timezone
from google.cloud.exceptions import NotFound
from google.cloud.storage.blob import Blob

from django_gcp.storage import gcloud

UTC = ZoneInfo("UTC")


class NonSeekableContentFile(ContentFile):
    def open(self, mode=None):
        return self

    def seekable(self):
        return False

    def seek(self, pos, whence=0):
        raise AttributeError()


class GCloudTestCase(TestCase):
    def setUp(self):
        self.client_patcher = mock.patch("django_gcp.storage.gcloud.Client")
        self.client_patcher.start()

    def tearDown(self):
        self.client_patcher.stop()


class GCloudStorageTests(GCloudTestCase):
    def setUp(self):
        super().setUp()
        self.bucket_name = "test-media"
        self.filename = "test_file.txt"
        self.storage = gcloud.GoogleCloudStorage(store_key="media", bucket_name=self.bucket_name)

    def test_open_read(self):
        """
        Test opening a file and reading from it
        """
        data = b"This is some test read data."

        f = self.storage.open(self.filename)
        self.storage._client.bucket.assert_called_with(self.bucket_name)
        self.storage._bucket.get_blob.assert_called_with(self.filename)

        f.blob.download_to_file = lambda tmpfile: tmpfile.write(data)
        self.assertEqual(f.read(), data)

    def test_open_read_num_bytes(self):
        data = b"This is some test read data."
        num_bytes = 10

        f = self.storage.open(self.filename)
        self.storage._client.bucket.assert_called_with(self.bucket_name)
        self.storage._bucket.get_blob.assert_called_with(self.filename)

        f.blob.download_to_file = lambda tmpfile: tmpfile.write(data)
        self.assertEqual(f.read(num_bytes), data[0:num_bytes])

    def test_open_read_nonexistent(self):
        self.storage._bucket = mock.MagicMock()
        self.storage._bucket.get_blob.return_value = None

        self.assertRaises(FileNotFoundError, self.storage.open, self.filename)
        self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_open_read_nonexistent_unicode(self):
        filename = "ủⓝï℅ⅆℇ.txt"

        self.storage._bucket = mock.MagicMock()
        self.storage._bucket.get_blob.return_value = None

        self.assertRaises(FileNotFoundError, self.storage.open, filename)

    @mock.patch("django_gcp.storage.gcloud.Blob")
    def test_open_write(self, MockBlob):
        """
        Test opening a file and writing to it
        """
        with override_settings(GCP_STORAGE_MEDIA={"bucket_name": self.bucket_name, "default_acl": "projectPrivate"}):
            data = "This is some test write data."

            # Simulate the file not existing before the write
            self.storage._bucket = mock.MagicMock()
            self.storage._bucket.get_blob.return_value = None

            f = self.storage.open(self.filename, "wb")
            MockBlob.assert_called_with(self.filename, self.storage._bucket, chunk_size=None)

            f.write(data)
            tmpfile = f._file
            # File data is not actually written until close(), so do that.
            f.close()

            MockBlob().upload_from_file.assert_called_with(
                tmpfile,
                rewind=True,
                content_type=mimetypes.guess_type(self.filename)[0],
                predefined_acl="projectPrivate",
            )

    def test_save(self):
        data = "This is some test content."
        content = ContentFile(data)

        self.storage.save(self.filename, content)

        self.storage._client.bucket.assert_called_with(self.bucket_name)
        self.storage._bucket.get_blob().upload_from_file.assert_called_with(
            content,
            rewind=True,
            size=len(data),
            content_type=mimetypes.guess_type(self.filename)[0],
            predefined_acl=None,
        )

    def test_save2(self):
        data = "This is some test ủⓝï℅ⅆℇ content."
        filename = "ủⓝï℅ⅆℇ.txt"
        content = ContentFile(data)

        self.storage.save(filename, content)

        self.storage._client.bucket.assert_called_with(self.bucket_name)
        self.storage._bucket.get_blob().upload_from_file.assert_called_with(
            content, rewind=True, size=len(data), content_type=mimetypes.guess_type(filename)[0], predefined_acl=None
        )

    def test_save_with_default_acl(self):
        data = "This is some test ủⓝï℅ⅆℇ content."
        filename = "ủⓝï℅ⅆℇ.txt"
        content = ContentFile(data)

        with override_settings(GCP_STORAGE_MEDIA={"bucket_name": self.bucket_name, "default_acl": "publicRead"}):
            self.storage.save(filename, content)
            self.storage._client.bucket.assert_called_with(self.bucket_name)
            self.storage._bucket.get_blob().upload_from_file.assert_called_with(
                content,
                rewind=True,
                size=len(data),
                content_type=mimetypes.guess_type(filename)[0],
                predefined_acl="publicRead",
            )

    def test_delete(self):
        self.storage.delete(self.filename)

        self.storage._client.bucket.assert_called_with(self.bucket_name)
        self.storage._bucket.delete_blob.assert_called_with(self.filename)

    def test_exists(self):
        self.storage._bucket = mock.MagicMock()
        self.assertTrue(self.storage.exists(self.filename))
        self.storage._bucket.get_blob.assert_called_with(self.filename)

        self.storage._bucket.reset_mock()
        self.storage._bucket.get_blob.return_value = None
        self.assertFalse(self.storage.exists(self.filename))
        self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_exists_no_bucket(self):
        # exists('') should return False if the bucket doesn't exist
        self.storage._client = mock.MagicMock()
        self.storage._client.get_bucket.side_effect = NotFound("dang")
        self.assertFalse(self.storage.exists(""))

    def test_exists_bucket(self):
        # exists('') should return True if the bucket exists
        self.assertTrue(self.storage.exists(""))

    def test_listdir(self):
        file_names = ["some/path/1.txt", "2.txt", "other/path/3.txt", "4.txt"]
        subdir = ""

        self.storage._bucket = mock.MagicMock()
        blobs, prefixes = [], []
        for name in file_names:
            directory = name.rsplit("/", 1)[0] + "/" if "/" in name else ""
            if directory == subdir:
                blob = mock.MagicMock(spec=Blob)
                blob.name = name.split("/")[-1]
                blobs.append(blob)
            else:
                prefixes.append(directory.split("/")[0] + "/")

        return_value = mock.MagicMock()
        return_value.__iter__ = mock.MagicMock(return_value=iter(blobs))
        return_value.prefixes = prefixes
        self.storage._bucket.list_blobs.return_value = return_value

        dirs, files = self.storage.listdir("")

        self.assertEqual(len(dirs), 2)
        for directory in ["some", "other"]:
            self.assertTrue(directory in dirs, f' "{directory}" not in directory list "{dirs}".')

        self.assertEqual(len(files), 2)
        for filename in ["2.txt", "4.txt"]:
            self.assertTrue(filename in files, f' "{filename}" not in file list "{files}".')

    def test_listdir_subdir(self):
        file_names = ["some/path/1.txt", "some/2.txt"]
        subdir = "some/"

        self.storage._bucket = mock.MagicMock()
        blobs, prefixes = [], []
        for name in file_names:
            directory = name.rsplit("/", 1)[0] + "/"
            if directory == subdir:
                blob = mock.MagicMock(spec=Blob)
                blob.name = name.split("/")[-1]
                blobs.append(blob)
            else:
                prefixes.append(directory.split(subdir)[1])

        return_value = mock.MagicMock()
        return_value.__iter__ = mock.MagicMock(return_value=iter(blobs))
        return_value.prefixes = prefixes
        self.storage._bucket.list_blobs.return_value = return_value

        dirs, files = self.storage.listdir(subdir)

        self.assertEqual(len(dirs), 1)
        self.assertTrue("path" in dirs, f' "path" not in directory list "{dirs}".')

        self.assertEqual(len(files), 1)
        self.assertTrue("2.txt" in files, f' "2.txt" not in files list "{files}".')

    def test_size(self):
        size = 1234

        self.storage._bucket = mock.MagicMock()
        blob = mock.MagicMock()
        blob.size = size
        self.storage._bucket.get_blob.return_value = blob

        self.assertEqual(self.storage.size(self.filename), size)
        self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_size_no_file(self):
        self.storage._bucket = mock.MagicMock()
        self.storage._bucket.get_blob.return_value = None

        self.assertRaises(NotFound, self.storage.size, self.filename)

    def test_modified_time(self):
        naive_date = datetime(2017, 1, 2, 3, 4, 5, 678)

        aware_date = timezone.make_aware(naive_date, UTC)

        self.storage._bucket = mock.MagicMock()
        blob = mock.MagicMock()
        blob.updated = aware_date
        self.storage._bucket.get_blob.return_value = blob

        with self.settings(TIME_ZONE="UTC"):
            mt = self.storage.modified_time(self.filename)
            self.assertTrue(timezone.is_naive(mt))
            self.assertEqual(mt, naive_date)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_get_modified_time(self):
        naive_date = datetime(2017, 1, 2, 3, 4, 5, 678)
        aware_date = timezone.make_aware(naive_date, UTC)

        self.storage._bucket = mock.MagicMock()
        blob = mock.MagicMock()
        blob.updated = aware_date
        self.storage._bucket.get_blob.return_value = blob

        with self.settings(TIME_ZONE="America/Montreal", USE_TZ=False):
            mt = self.storage.get_modified_time(self.filename)
            self.assertTrue(timezone.is_naive(mt))
            naive_date_montreal = timezone.make_naive(aware_date)
            self.assertEqual(mt, naive_date_montreal)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

        with self.settings(TIME_ZONE="America/Montreal", USE_TZ=True):
            mt = self.storage.get_modified_time(self.filename)
            self.assertTrue(timezone.is_aware(mt))
            self.assertEqual(mt, aware_date)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_get_created_time(self):
        naive_date = datetime(2017, 1, 2, 3, 4, 5, 678)
        aware_date = timezone.make_aware(naive_date, UTC)

        self.storage._bucket = mock.MagicMock()
        blob = mock.MagicMock()
        blob.time_created = aware_date
        self.storage._bucket.get_blob.return_value = blob

        with self.settings(TIME_ZONE="America/Montreal", USE_TZ=False):
            mt = self.storage.get_created_time(self.filename)
            self.assertTrue(timezone.is_naive(mt))
            naive_date_montreal = timezone.make_naive(aware_date)
            self.assertEqual(mt, naive_date_montreal)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

        with self.settings(TIME_ZONE="America/Montreal", USE_TZ=True):
            mt = self.storage.get_created_time(self.filename)
            self.assertTrue(timezone.is_aware(mt))
            self.assertEqual(mt, aware_date)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_modified_time_no_file(self):
        self.storage._bucket = mock.MagicMock()
        self.storage._bucket.get_blob.return_value = None

        self.assertRaises(NotFound, self.storage.modified_time, self.filename)

    def test_url_public_object(self):
        url = f"https://example.com/mah-bukkit/{self.filename}"

        with override_settings(GCP_STORAGE_MEDIA={"bucket_name": self.bucket_name, "default_acl": "publicRead"}):
            self.storage._bucket = mock.MagicMock()
            blob = mock.MagicMock()
            blob.public_url = url
            blob.generate_signed_url = "not called"
            self.storage._bucket.blob.return_value = blob

            self.assertEqual(self.storage.url(self.filename), url)
            self.storage._bucket.blob.assert_called_with(self.filename)

    def test_url_not_public_file(self):
        secret_filename = "secret_file.txt"
        self.storage._bucket = mock.MagicMock()
        blob = mock.MagicMock()
        generate_signed_url = mock.MagicMock(return_value="http://signed_url")
        blob.public_url = "http://this_is_public_url"
        blob.generate_signed_url = generate_signed_url
        self.storage._bucket.blob.return_value = blob

        url = self.storage.url(secret_filename)
        self.storage._bucket.blob.assert_called_with(secret_filename)
        self.assertEqual(url, "http://signed_url")
        blob.generate_signed_url.assert_called_with(expiration=timedelta(seconds=86400), version="v4")

    def test_url_not_public_file_with_custom_expires(self):
        expiration = timedelta(seconds=3600)
        with override_settings(GCP_STORAGE_MEDIA={"bucket_name": self.bucket_name, "expiration": expiration}):
            secret_filename = "secret_file.txt"
            self.storage._bucket = mock.MagicMock()
            blob = mock.MagicMock()
            generate_signed_url = mock.MagicMock(return_value="http://signed_url")
            blob.generate_signed_url = generate_signed_url
            self.storage._bucket.blob.return_value = blob
            url = self.storage.url(secret_filename)
            self.storage._bucket.blob.assert_called_with(secret_filename)
            self.assertEqual(url, "http://signed_url")
            blob.generate_signed_url.assert_called_with(expiration=expiration, version="v4")

    def test_custom_endpoint(self):
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "custom_endpoint": "https://example.com",
                "default_acl": "publicRead",
            }
        ):
            url = f"{self.storage.settings.custom_endpoint}/{self.filename}"
            self.assertEqual(self.storage.url(self.filename), url)

        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "custom_endpoint": "https://example.com",
                "default_acl": "projectPrivate",
            }
        ):
            bucket_name = "hyacinth"
            self.storage._bucket = mock.MagicMock()
            blob = mock.MagicMock()
            generate_signed_url = mock.MagicMock()
            blob.bucket = mock.MagicMock()
            type(blob.bucket).name = mock.PropertyMock(return_value=bucket_name)
            blob.generate_signed_url = generate_signed_url
            self.storage._bucket.blob.return_value = blob
            self.storage.url(self.filename)
            blob.generate_signed_url.assert_called_with(
                bucket_bound_hostname=self.storage.settings.custom_endpoint,
                expiration=timedelta(seconds=86400),
                version="v4",
            )

    def test_get_available_name(self):
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "file_overwrite": True,
            }
        ):
            self.assertEqual(self.storage.get_available_name(self.filename), self.filename)
            self.storage._bucket = mock.MagicMock()
            self.storage._bucket.get_blob.return_value = None
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "file_overwrite": False,
            }
        ):
            self.assertEqual(self.storage.get_available_name(self.filename), self.filename)
            self.storage._bucket.get_blob.assert_called_with(self.filename)

    def test_get_available_name_unicode(self):
        filename = "ủⓝï℅ⅆℇ.txt"
        self.assertEqual(self.storage.get_available_name(filename), filename)

    def test_cache_control(self):
        data = "This is some test content."
        filename = "cache_control_file.txt"
        content = ContentFile(data)
        cache_control = "public, max-age=604800"
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "file_overwrite": True,
                "object_parameters": {"cache_control": cache_control},
            }
        ):
            self.storage.save(filename, content)
            bucket = self.storage.client.bucket(self.bucket_name)
            blob = bucket.get_blob(filename)
            self.assertEqual(blob.cache_control, cache_control)

    def test_storage_save_gzipped(self):
        """
        Test saving a gzipped file
        """
        name = "test_storage_save.gz"
        content = ContentFile("I am gzip'd")
        self.storage.save(name, content)
        obj = self.storage._bucket.get_blob()
        obj.upload_from_file.assert_called_with(mock.ANY, rewind=True, size=11, predefined_acl=None, content_type=None)

    def test_storage_save_gzipped_non_seekable(self):
        """
        Test saving a gzipped file
        """
        name = "test_storage_save.gz"
        content = NonSeekableContentFile("I am gzip'd")
        self.storage.save(name, content)
        obj = self.storage._bucket.get_blob()
        obj.upload_from_file.assert_called_with(mock.ANY, rewind=True, size=11, predefined_acl=None, content_type=None)

    def test_storage_save_gzip(self):
        """
        Test saving a file with gzip enabled.
        """
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "gzip": True,
            }
        ):
            name = "test_storage_save.css"
            content = ContentFile("I should be gzip'd")
            self.storage.save(name, content)
            self.storage._client.bucket.assert_called_with(self.bucket_name)
            obj = self.storage._bucket.get_blob()
            self.assertEqual(obj.content_encoding, "gzip")
            obj.upload_from_file.assert_called_with(
                mock.ANY,
                rewind=True,
                size=None,
                predefined_acl=None,
                content_type="text/css",
            )
            args, _ = obj.upload_from_file.call_args
            content = args[0]
            zfile = gzip.GzipFile(mode="rb", fileobj=content)
            self.assertEqual(zfile.read(), b"I should be gzip'd")

    def test_storage_save_gzip_twice(self):
        """
        Test saving the same file content twice with gzip enabled.
        """
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": self.bucket_name,
                "gzip": True,
            }
        ):
            name = "test_storage_save.css"
            content = ContentFile("I should be gzip'd")

            # When
            self.storage.save(name, content)
            self.storage.save("test_storage_save_2.css", content)

            # Then
            self.storage._client.bucket.assert_called_with(self.bucket_name)
            obj = self.storage._bucket.get_blob()
            self.assertEqual(obj.content_encoding, "gzip")
            obj.upload_from_file.assert_called_with(
                mock.ANY,
                rewind=True,
                size=None,
                predefined_acl=None,
                content_type="text/css",
            )
            args, _ = obj.upload_from_file.call_args
            content = args[0]
            zfile = gzip.GzipFile(mode="rb", fileobj=content)
            self.assertEqual(zfile.read(), b"I should be gzip'd")

    def test_compress_content_len(self):
        """
        Test that file returned by _compress_content() is readable.
        """
        self.storage.gzip = True
        content = ContentFile("I should be gzip'd")
        content = self.storage._compress_content(content)
        self.assertTrue(len(content.read()) > 0)

    def test_location_leading_slash(self):
        msg = "'location' option in GCP_STORAGE_ cannot begin with a leading slash. Found '/'. Use '' instead."
        with self.assertRaises(ImproperlyConfigured, msg=msg):
            with override_settings(
                GCP_STORAGE_MEDIA={
                    "bucket_name": self.bucket_name,
                    "location": "/",
                }
            ):
                self.storage.settings.check()


class GCloudStorageClassTests(GCloudTestCase):
    def test_override_settings(self):
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": "test_media",
                "location": "foo1",
            }
        ):
            storage = gcloud.GoogleCloudStorage()
            self.assertEqual(storage.settings.location, "foo1")
        with override_settings(
            GCP_STORAGE_MEDIA={
                "bucket_name": "test_media",
                "location": "foo2",
            }
        ):
            storage = gcloud.GoogleCloudStorage()
            self.assertEqual(storage.settings.location, "foo2")

    def test_override_init_argument(self):
        storage = gcloud.GoogleCloudStorage(location="foo1")
        self.assertEqual(storage.settings.location, "foo1")
        storage = gcloud.GoogleCloudStorage(location="foo2")
        self.assertEqual(storage.settings.location, "foo2")

    def test_media_storage_instantiation(self):
        storage = gcloud.GoogleCloudMediaStorage()
        self.assertEqual(storage.settings.bucket_name, "example-media-assets")

    def test_static_storage_instantiation(self):
        storage = gcloud.GoogleCloudStaticStorage()
        self.assertEqual(storage.settings.bucket_name, "example-static-assets")

    def test_instiantiation_with_store_key_raises_exception(self):
        with self.assertRaises(ValueError):
            gcloud.GoogleCloudMediaStorage(store_key="not-media")

        with self.assertRaises(ValueError):
            gcloud.GoogleCloudStaticStorage(store_key="not-static")



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

import json
from unittest.mock import patch

from django.test import SimpleTestCase
from django.urls import reverse

from django_gcp.events.utils import make_pubsub_message
from django_gcp.tasks._pilot.mocker import patch_auth

from .test_events_utils import DEFAULT_SUBSCRIPTION


class ExampleAppViewTest(SimpleTestCase):
    def test_task_pushed_to_queue(self):
        url = reverse("enqueue-on-demand")

        with self.settings(GCP_TASKS_DOMAIN="https://the-domain.com"):
            with patch("django_gcp.tasks._pilot.tasks.CloudTasks.push") as patched_push:
                with patch_auth():
                    response = self.client.post(path=url, content_type="application/json")

            self.assertEqual(201, response.status_code)
            expected_call = dict(
                queue_name="example-primary",
                url="https://the-domain.com/example-django-gcp/tasks/MyOnDemandTask",
                payload='{"a": 1}',
            )
            patched_push.assert_called_once_with(**expected_call)


class TasksViewTest(SimpleTestCase):
    def test_on_demand_task_run_method_called_with_data(self):
        url = reverse("gcp-tasks", args=["MyOnDemandTask"])
        data = {"a": 1}
        patch_response = {"b": 2}
        with patch("tests.server.example.tasks.MyOnDemandTask.run", return_value=patch_response) as patched_run:
            response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(200, response.status_code)
        self.assertEqual({"result": patch_response}, response.json())
        patched_run.assert_called_once_with(**data)

    def test_periodic_task_run_method_called_with_data(self):
        url = reverse("gcp-tasks", args=["MyPeriodicTask"])
        data = {"a": 1}
        patch_response = {"b": 2}
        with patch("tests.server.example.tasks.MyPeriodicTask.run", return_value=patch_response) as patched_run:
            response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(200, response.status_code)
        self.assertEqual({"result": patch_response}, response.json())
        patched_run.assert_called_once_with(**data)

    def test_invalid_task_name(self):
        url = reverse("gcp-tasks", args=["NotAValidTaskName"])
        data = {"a": 1}
        response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(404, response.status_code)
        self.assertIn("error", response.json())
        self.assertEqual("Task NotAValidTaskName not found", response.json()["error"])

    def test_subscriber_task_run_method_called_with_data(self):
        url = reverse("gcp-subscriber-tasks", args=["MySubscriberTask"])

        msg = make_pubsub_message({"a": 1}, DEFAULT_SUBSCRIPTION)
        with patch("tests.server.example.tasks.MySubscriberTask.run", return_value=None) as patched_run:
            response = self.client.post(path=url, data=msg, content_type="application/json")

        self.assertEqual(200, response.status_code)
        self.assertEqual({"result": None}, response.json())
        patched_run.assert_called_once()

    @patch("django_gcp.logs.error_reporting.GoogleErrorReportingHandler.emit")
    def test_failing_on_demand_task_returns_error(self, patched_emit):
        url = reverse("gcp-tasks", args=["FailingOnDemandTask"])
        data = {"a": 1}
        response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(500, response.status_code)
        self.assertIn("error", response.json())
        patched_emit.assert_called()



# pylint: disable=missing-docstring

import datetime

from django.core.exceptions import SuspiciousFileOperation
from django.test import TestCase

from django_gcp.storage import utils


class CleanNameTests(TestCase):
    def test_clean_name(self):
        """
        Test the base case of clean_name
        """
        path = utils.clean_name("path/to/somewhere")
        self.assertEqual(path, "path/to/somewhere")

    def test_clean_name_normalize(self):
        """
        Test the normalization of clean_name
        """
        path = utils.clean_name("path/to/../somewhere")
        self.assertEqual(path, "path/somewhere")

    def test_clean_name_trailing_slash(self):
        """
        Test the clean_name when the path has a trailing slash
        """
        path = utils.clean_name("path/to/somewhere/")
        self.assertEqual(path, "path/to/somewhere/")

    def test_clean_name_windows(self):
        """
        Test the clean_name when the path has a trailing slash
        """
        path = utils.clean_name("path\\to\\somewhere")
        self.assertEqual(path, "path/to/somewhere")


class SafeJoinTest(TestCase):
    def test_normal(self):
        path = utils.safe_join("", "path/to/somewhere", "other", "path/to/somewhere")
        self.assertEqual(path, "path/to/somewhere/other/path/to/somewhere")

    def test_with_dot(self):
        path = utils.safe_join("", "path/./somewhere/../other", "..", ".", "to/./somewhere")
        self.assertEqual(path, "path/to/somewhere")

    def test_with_only_dot(self):
        path = utils.safe_join("", ".")
        self.assertEqual(path, "")

    def test_base_url(self):
        path = utils.safe_join("base_url", "path/to/somewhere")
        self.assertEqual(path, "base_url/path/to/somewhere")

    def test_base_url_with_slash(self):
        path = utils.safe_join("base_url/", "path/to/somewhere")
        self.assertEqual(path, "base_url/path/to/somewhere")

    def test_suspicious_operation(self):
        with self.assertRaises(ValueError):
            utils.safe_join("base", "../../../../../../../etc/passwd")
        with self.assertRaises(ValueError):
            utils.safe_join("base", "/etc/passwd")

    def test_trailing_slash(self):
        """
        Test safe_join with paths that end with a trailing slash.
        """
        path = utils.safe_join("base_url/", "path/to/somewhere/")
        self.assertEqual(path, "base_url/path/to/somewhere/")

    def test_trailing_slash_multi(self):
        """
        Test safe_join with multiple paths that end with a trailing slash.
        """
        path = utils.safe_join("base_url/", "path/to/", "somewhere/")
        self.assertEqual(path, "base_url/path/to/somewhere/")

    def test_datetime_isoformat(self):
        dt = datetime.datetime(2017, 5, 19, 14, 45, 37, 123456)
        path = utils.safe_join("base_url", dt.isoformat())
        self.assertEqual(path, "base_url/2017-05-19T14:45:37.123456")

    def test_join_empty_string(self):
        path = utils.safe_join("base_url", "")
        self.assertEqual(path, "base_url/")

    def test_with_base_url_and_dot(self):
        path = utils.safe_join("base_url", ".")
        self.assertEqual(path, "base_url/")

    def test_with_base_url_and_dot_and_path_and_slash(self):
        path = utils.safe_join("base_url", ".", "path/to/", ".")
        self.assertEqual(path, "base_url/path/to/")

    def test_join_nothing(self):
        path = utils.safe_join("")
        self.assertEqual(path, "")

    def test_with_base_url_join_nothing(self):
        path = utils.safe_join("base_url")
        self.assertEqual(path, "base_url/")


class TestGetAvailableOverwriteName(TestCase):
    def test_maxlength_is_none(self):
        name = "superlong/file/with/path.txt"
        self.assertEqual(utils.get_available_overwrite_name(name, None), name)

    def test_maxlength_equals_name(self):
        name = "parent/child.txt"
        self.assertEqual(utils.get_available_overwrite_name(name, len(name)), name)

    def test_maxlength_is_greater_than_name(self):
        name = "parent/child.txt"
        self.assertEqual(utils.get_available_overwrite_name(name, len(name) + 1), name)

    def test_maxlength_less_than_name(self):
        name = "parent/child.txt"
        self.assertEqual(utils.get_available_overwrite_name(name, len(name) - 1), "parent/chil.txt")

    def test_truncates_away_filename_raises(self):
        name = "parent/child.txt"
        with self.assertRaises(SuspiciousFileOperation):
            utils.get_available_overwrite_name(name, len(name) - 5)



# pylint: disable=missing-docstring

from datetime import date
import os
import tempfile
from uuid import uuid4

from django.test import SimpleTestCase, TestCase
from google.cloud import storage

from django_gcp.exceptions import AttemptedOverwriteError, MissingBlobError
from django_gcp.storage.blob_utils import get_blob
from django_gcp.storage.operations import copy_blob, delete_blob, get_generations, uploaded_blob
from tests.server.example.models import ExampleBlobFieldModel


class StorageOperationsMixin:
    """A mixin allowing TestCase to do full integration tests on GCS, with storage operations"""

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)
        self.storage_client = storage.Client()
        # Allows us to avoid cleanup overhead by having a sweeper function trigger
        # periodically to clear up test artefacts older than today,
        # and allows us to avoid race conditions by having a per-test UID.
        uid = str(uuid4()).replace("-", "")[0:8]
        today = str(date.today())
        self.test_path = f"test/{today}--{uid}/"

        # Allows us to choose either a regular or versioned bucket
        self.bucket = self.storage_client.bucket("example-media-assets")
        self.versioned_bucket = self.storage_client.bucket("example-extra-versioned-assets")

    def _prefix_blob_name(self, blob_name):
        """Adds the test prefix to a blob name"""
        return os.path.join(self.test_path, blob_name)

    def _create_test_blob(self, bucket, blob_name, content=""):
        """Create a plain text file and upload it to a blob under the test path, for test purposes"""
        blob = storage.Blob(name=blob_name, bucket=bucket)
        blob.upload_from_string(content)
        return blob

    def _create_temporary_blob(self, bucket, tmp_path=None, content=""):
        """Create a plain text file and upload it to a blob at a temporary path"""
        tmp_path = tmp_path or f"_tmp/{uuid4()}"
        blob = storage.Blob(name=tmp_path, bucket=bucket)
        blob.upload_from_string(content)
        return blob

    def _copy(self, bucket, test_name, **extras):
        source_blob_name = self._prefix_blob_name(f"{test_name}.source.txt")
        destination_blob_name = self._prefix_blob_name(f"{test_name}.destination.txt")

        source_blob = self._create_test_blob(bucket, source_blob_name)
        destination_blob = copy_blob(bucket, source_blob_name, bucket, destination_blob_name, **extras)
        return source_blob, destination_blob


class TestStorageOperations(StorageOperationsMixin, SimpleTestCase):
    """This class tests the StorageOperations module, and comprises a full
    integration test where operations are actually executed on GCS.
    """

    def test_copy_blob(self):
        """Test that a blob can be copied"""

        source_blob, destination_blob = self._copy(self.bucket, "test_copy_blob")

        # Ensure source and destination present
        self.assertTrue(source_blob.exists())
        self.assertTrue(destination_blob.exists())

    def test_move_blob(self):
        """Test that a blob can be moved"""
        source_blob, destination_blob = self._copy(self.bucket, "test_copy_blob", move=True)

        # Ensure source gone and destination present
        self.assertFalse(source_blob.exists())
        self.assertTrue(destination_blob.exists())

    def test_copy_blob_allows_overwrite(self):
        """Test that a blob can be overwritten if allowed"""

        source_blob_name = self._prefix_blob_name("test_copy_blob_allows_overwrite.to_overwrite.txt")
        destination_blob_name = self._prefix_blob_name("test_copy_blob_allows_overwrite.to_be_overwritten.txt")

        self._create_test_blob(self.bucket, source_blob_name, "to overwrite")
        self._create_test_blob(self.bucket, destination_blob_name, "to be overwritten")
        destination_blob = copy_blob(self.bucket, source_blob_name, self.bucket, destination_blob_name, overwrite=True)

        with destination_blob.open("r") as fp:
            content = fp.read()

        self.assertEqual(content, "to overwrite")

    def test_copy_blob_prevents_overwrite(self):
        """Test that overwrite is prevented by default"""
        source_blob_name = self._prefix_blob_name("test_copy_blob_prevents_overwrite.to_overwrite.txt")
        destination_blob_name = self._prefix_blob_name("test_copy_blob_prevents_overwrite.to_be_overwritten.txt")

        self._create_test_blob(self.bucket, source_blob_name, "to overwrite")
        self._create_test_blob(self.bucket, destination_blob_name, "to be overwritten")

        with self.assertRaises(AttemptedOverwriteError):
            copy_blob(self.bucket, source_blob_name, self.bucket, destination_blob_name)

    def test_overwrite_increments_generation(self):
        """Test that overwriting a blob in a versioned store creates a new generation"""

        source_blob_name = self._prefix_blob_name("test_overwrite_increments_generation.to_overwrite.txt")
        destination_blob_name = self._prefix_blob_name("test_overwrite_increments_generation.to_be_overwritten.txt")

        self._create_test_blob(self.versioned_bucket, source_blob_name, "to overwrite")
        og_destination_blob = self._create_test_blob(self.versioned_bucket, destination_blob_name, "to be overwritten")
        destination_blob = copy_blob(
            self.versioned_bucket,
            source_blob_name,
            self.versioned_bucket,
            destination_blob_name,
            overwrite=True,
        )

        with destination_blob.open("r") as fp:
            content = fp.read()

        self.assertEqual(content, "to overwrite")
        self.assertNotEqual(og_destination_blob.generation, destination_blob.generation)

    def test_copy_blob_handles_missing(self):
        """Test that overwrite is prevented by default"""

        source_blob_name = self._prefix_blob_name("test_copy_blob_prevents_overwrite.to_overwrite.txt")
        destination_blob_name = self._prefix_blob_name("test_copy_blob_prevents_overwrite.to_be_overwritten.txt")

        with self.assertRaises(MissingBlobError):
            copy_blob(self.bucket, source_blob_name, self.bucket, destination_blob_name)

    def test_delete_blob(self):
        """Test that deletion works"""
        blob_name = self._prefix_blob_name("test_delete_blob.to_delete.txt")
        blob = self._create_test_blob(self.bucket, blob_name)
        self.assertTrue(blob.exists())
        delete_blob(self.bucket, blob_name)
        self.assertFalse(blob.exists())

    def test_delete_versioned_blob(self):
        """Test that deletion works as intended on versioned objects
        In versioned objects, the blob should record when it was deleted.
        Only if passed a specific generation number should the blob actually be removed
        """
        # Prepare a blob for deletion
        blob_name = self._prefix_blob_name("test_delete_versioned_blob.to_delete.txt")
        blob = self._create_test_blob(self.versioned_bucket, blob_name)
        self.assertTrue(blob.exists())
        self.assertIsNone(blob.time_deleted)

        # Delete the versioned blob, without specifying generation...
        delete_blob(self.versioned_bucket, blob_name)
        blob.reload()
        self.assertIsNotNone(blob.time_deleted)

        # Repeat, specifying generation to ensure removal of the actual object...
        delete_blob(self.versioned_bucket, blob_name, generation=blob.generation)
        self.assertFalse(blob.exists())

    def test_delete_missing_blob(self):
        """Test that deletion works in the event of a missing blob"""
        blob_name = self._prefix_blob_name("test_delete_missing_blob.txt")
        # Don't create the blob in cloud, just locally so we can use the exists() method
        blob = storage.Blob(name=blob_name, bucket=self.bucket)
        self.assertFalse(blob.exists())
        delete_blob(self.bucket, blob_name, ignore_missing=True)
        self.assertFalse(blob.exists())
        with self.assertRaises(MissingBlobError):
            delete_blob(self.bucket, blob_name)

    def test_get_generations(self):
        blob_name = self._prefix_blob_name("test_get_generations.txt")
        first_version = self._create_test_blob(
            self.versioned_bucket,
            blob_name,
            content="first generation",
        )
        second_version = self._create_test_blob(
            self.versioned_bucket,
            blob_name,
            content="second generation",
        )
        blobs = get_generations(self.versioned_bucket, blob_name)
        generations = [blob.generation for blob in blobs]
        self.assertEqual(len(generations), 2)
        self.assertIn(first_version.generation, generations)
        self.assertIn(second_version.generation, generations)


class TestStorageOperationsWithDatabase(StorageOperationsMixin, TestCase):
    def test_uploaded_blob(self):
        """Ensure that test_uploaded_blob will upload a blob and leave it there"""

        # Create a local file to upload
        with tempfile.TemporaryDirectory() as tmpdir:
            local_path = os.path.join(tmpdir, "test_file.txt")
            with open(local_path, "w"):
                pass  # Create an empty file

            self.assertTrue(os.path.exists(local_path))

            # Test with no destination path
            instance = ExampleBlobFieldModel()
            field_name = "blob"
            with uploaded_blob(instance, field_name, local_path) as value:
                instance.blob = value
                instance.save()
            assert get_blob(instance, "blob").exists()

            # Test with a file-like
            instance = ExampleBlobFieldModel()
            field_name = "blob"
            with open(local_path, "rb") as local_fp:
                with uploaded_blob(instance, field_name, local_fp) as value:
                    instance.blob = value
                    instance.save()
            assert get_blob(instance, "blob").exists()

            # Test raises exception on not implemented yet
            with self.assertRaises(NotImplementedError):
                instance2 = ExampleBlobFieldModel()
                with uploaded_blob(instance2, field_name, local_path, delete_on_exit=True) as value:
                    instance2.blob = value
                    instance2.save()



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods
# pylint: disable=no-member
from unittest.mock import patch
from uuid import uuid4

from django import forms
from django.contrib.auth.models import User
from django.db import transaction
from django.test import Client, TestCase, TransactionTestCase, override_settings

from django_gcp.exceptions import MissingBlobError
from django_gcp.storage.blob_utils import get_blob
from tests.server.example.models import (
    ExampleBlankBlobFieldModel,
    ExampleBlobFieldModel,
    ExampleCallbackBlobFieldModel,
    ExampleUpdateAttributesBlobFieldModel,
)

from ._utils import get_admin_add_view_url, get_admin_change_view_url
from .test_storage_operations import StorageOperationsMixin


class BlobForm(forms.ModelForm):
    """Dummy form for testing modeladmin"""

    class Meta:
        model = ExampleBlobFieldModel
        fields = ["blob"]


class BlankBlobForm(forms.ModelForm):
    """Dummy form for testing modeladmin"""

    class Meta:
        model = ExampleBlankBlobFieldModel
        fields = ["blob"]


class CallbackBlobForm(forms.ModelForm):
    """Dummy form for testing modeladmin"""

    class Meta:
        model = ExampleCallbackBlobFieldModel
        fields = ["blob"]


class UpdateAttributesBlobForm(forms.ModelForm):
    """Dummy form for testing modeladmin"""

    class Meta:
        model = ExampleUpdateAttributesBlobFieldModel
        fields = ["blob"]


class BlobModelFactoryMixin:
    def _create(self, Model, name=None, content="", **create_kwargs):
        """Through the ORM, we may need to create blobs directly at the destination"""
        with override_settings(GCP_STORAGE_ALLOW_PATH_OVERRIDE=True):
            if name is not None:
                blob_name = self._prefix_blob_name(name)
                self._create_test_blob(self.bucket, blob_name, content)
                blob = {"path": blob_name}
            else:
                blob = None
            obj = Model.objects.create(blob=blob, **create_kwargs)
            obj.save()
        return obj


def get_destination_path_for_test(
    instance,
    original_name,
    attributes,
    existing_path,
    temporary_path,
    allow_overwrite,
    bucket,
):
    """Returns a consistent destination path for a blob for test purposes"""
    category = f"{instance.category}/" if instance.category is not None else ""
    return f"{category}{original_name}", allow_overwrite


class TestBlobFieldAdmin(StorageOperationsMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        """Set up data for the whole test case (used within the outer transaction to increase speed)"""
        User.objects.create_superuser(username="superuser", password="secret", email="admin@example.com")

    def setUp(self, *args, **kwargs):
        """Log in the superuser"""
        super().setUp(*args, **kwargs)
        self.client = Client()
        self.client.login(username="superuser", password="secret")

    def test_add_view_loads_normally(self):
        response = self.client.get(get_admin_add_view_url(ExampleBlobFieldModel))
        self.assertEqual(response.status_code, 200)

    def test_add_view_has_presigned_url(self):
        """The add view must have a presigned URL available in the context for uploading to a temporary path"""
        response = self.client.get(get_admin_add_view_url(ExampleBlobFieldModel))
        self.assertEqual(response.status_code, 200)
        widget = response.context_data["adminform"].fields["blob"].widget
        self.assertTrue(hasattr(widget, "signed_ingress_url"))
        self.assertTrue(
            widget.signed_ingress_url.startswith("https://storage.googleapis.com/example-media-assets/_tmp")
        )

    def test_full_clean_executes_in_overridden_context(self):
        """Ensure that full_clean() can be legitimately called
        on a model while in an overridden context
        """

        blob_name = self._prefix_blob_name("test_full_clean_executes_in_overridden_context.txt")
        self._create_test_blob(self.bucket, blob_name, "")
        with override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True):
            obj = ExampleBlobFieldModel(blob={"path": blob_name})
            obj.full_clean()
            obj.save()

    @override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True)
    def test_change_view_loads_normally(self):
        """Ensure we can load a change view"""

        blob_name = self._prefix_blob_name("test_change_view_loads_normally.txt")
        self._create_test_blob(self.bucket, blob_name, "")
        obj = ExampleBlobFieldModel.objects.create(blob={"path": blob_name})

        # Assert that the view loads
        response = self.client.get(get_admin_change_view_url(obj))
        self.assertEqual(response.status_code, 200)

        # Assert the widget contents
        widget = response.context_data["adminform"].fields["blob"].widget
        self.assertTrue(hasattr(widget, "signed_ingress_url"))
        self.assertTrue(
            widget.signed_ingress_url.startswith("https://storage.googleapis.com/example-media-assets/_tmp")
        )


class TestBlobField(StorageOperationsMixin, TestCase):
    """Inherits from transaction test case, because we use an on_commit
    hook to move ingressed files once a database save has been made

    """

    def test_validates_name_is_present_on_add(self):
        """Ensure that a ValidationError is raised if no 'name' property is present in the blob data"""

        form = BlobForm(data={"blob": {"_tmp_path": "something"}})

        self.assertEqual(
            form.errors["blob"],
            [
                "Both `_tmp_path` and `name` properties must be present in data for ExampleBlobFieldModel.blob if ingressing a new blob."
            ],
        )

    def test_validates_path_not_present_on_add(self):
        """Ensure that a ValidationError is raised if no 'name' property is present in the blob data"""

        form = BlobForm(
            data={
                "blob": {
                    "name": "something",
                    "_tmp_path": "something",
                    "path": "something",
                }
            }
        )

        self.assertEqual(
            form.errors["blob"],
            ["You cannot specify a path directly"],
        )

    def test_validates_tmp_path_is_present_on_add(self):
        """Ensure that a ValidationError is raised if no 'name' property is present in the blob data"""

        form = BlobForm(data={"blob": {"name": "something"}})

        self.assertEqual(
            form.errors["blob"],
            [
                "Both `_tmp_path` and `name` properties must be present in data for ExampleBlobFieldModel.blob if ingressing a new blob."
            ],
        )

    def test_validates_raises_on_add_blank_dict(self):
        form = BlobForm(data={"blob": {}})
        self.assertEqual(
            form.errors["blob"],
            ["This field is required."],
        )

    def test_validates_raises_on_add_blank_string(self):
        form = BlobForm(data={"blob": ""})

        self.assertEqual(
            form.errors["blob"],
            ["This field is required."],
        )

    def test_validates_raises_on_add_blank_none(self):
        form = BlobForm(data={"blob": None})

        self.assertEqual(
            form.errors["blob"],
            ["This field is required."],
        )

    @override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True)
    def test_create_object_succeeds_with_overridden_path(self):
        """Through the ORM, we may need to create blobs directly at the destination"""

        with self.captureOnCommitCallbacks(execute=False) as callbacks:
            blob_name = self._prefix_blob_name("create_object_succeeds_with_overridden_path.txt")
            self._create_test_blob(self.bucket, blob_name)
            ExampleBlobFieldModel.objects.create(blob={"path": blob_name})

        # There should be no callback executed when the blob field value is overwritten directly
        self.assertEqual(len(callbacks), 0)
        count = ExampleBlobFieldModel.objects.count()
        self.assertEqual(count, 1)

    def test_create_object_fails_with_missing_blob(self):
        """Create an object but fail to copy the blob (because it's missing) then check that
        no database record was created"""
        with self.assertRaises(MissingBlobError):
            with self.captureOnCommitCallbacks(execute=True) as callbacks:
                ExampleBlobFieldModel.objects.create(
                    blob={"_tmp_path": f"_tmp/{uuid4()}.txt", "name": "missing_blob.txt"}
                )
            self.assertEqual(len(callbacks), 1)
            count = ExampleBlobFieldModel.objects.count()
            self.assertEqual(count, 0)


class TestBlankBlobField(BlobModelFactoryMixin, StorageOperationsMixin, TransactionTestCase):
    """Inherits from transaction test case, because we use an on_commit
    hook to move ingressed files once a database save has been made

    TODO REMOVE TRANSACTION TEST CASE
    as per https://code.djangoproject.com/ticket/30457
    ```
    with self.captureOnCommitCallbacks(execute=True) as callbacks:
        with transaction.atomic():
            transaction.on_commit(branch_1)
    ```
    """

    def test_validates_on_add_blank_dict(self):
        form = BlankBlobForm(data={"blob": {}})
        self.assertNotIn("blob", form.errors)

    def test_validates_on_add_blank_string(self):
        form = BlankBlobForm(data={"blob": ""})
        self.assertNotIn("blob", form.errors)

    def test_validates_on_add_blank_none(self):
        form = BlankBlobForm(data={"blob": None})
        self.assertNotIn("blob", form.errors)

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_update_valid_to_blank(self):
        # Create a valid instance
        name = self._prefix_blob_name("update_valid_to_blank.txt")
        tmp_blob = self._create_temporary_blob(self.bucket)
        form = BlankBlobForm(data={"blob": {"_tmp_path": tmp_blob.name, "name": name}})
        instance = form.save()

        # Create a form to change this instance and set the data blank
        form = BlankBlobForm(instance=instance, data={"blob": {}})
        self.assertTrue(form.is_valid())
        form.save()
        instance.refresh_from_db()
        self.assertIsNone(instance.blob)

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_update_blank_to_valid(self):
        # Create a blank instance
        form = BlankBlobForm(data={"blob": {}})
        instance = form.save()
        instance.refresh_from_db()

        name = self._prefix_blob_name("update_blank_to_valid.txt")
        tmp_blob = self._create_temporary_blob(self.bucket)

        # Create a form to change this instance and set the data to a real value
        form = BlankBlobForm(instance=instance, data={"blob": {"_tmp_path": tmp_blob.name, "name": name}})
        self.assertTrue(form.is_valid())
        form.save()
        instance.refresh_from_db()
        self.assertIsNotNone(instance.blob)
        self.assertIn("path", instance.blob)
        self.assertEqual(instance.blob["path"], name)

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_update_valid_to_valid(self):
        # Create a valid instance
        name = self._prefix_blob_name("update_valid_to_valid.txt")
        tmp_blob = self._create_temporary_blob(self.bucket)
        form = BlankBlobForm(data={"blob": {"_tmp_path": tmp_blob.name, "name": name}})
        instance = form.save()
        instance.refresh_from_db()

        new_name = self._prefix_blob_name("overwrite_update_valid_to_valid.txt")
        tmp_blob = self._create_temporary_blob(self.bucket)

        # Create a form to change this instance and set the data to a real value
        form = BlankBlobForm(
            instance=instance,
            data={"blob": {"_tmp_path": tmp_blob.name, "name": new_name}},
        )
        self.assertTrue(form.is_valid())
        form.save()
        instance.refresh_from_db()
        self.assertIsNotNone(instance.blob)
        self.assertIn("path", instance.blob)
        self.assertEqual(instance.blob["path"], new_name)

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_update_valid_unchanged(self):
        # Create a valid instance
        name = self._prefix_blob_name("update_valid_unchanged.txt")
        tmp_blob = self._create_temporary_blob(self.bucket)
        form = BlankBlobForm(data={"blob": {"_tmp_path": tmp_blob.name, "name": name}})
        instance = form.save()
        instance.refresh_from_db()
        self.assertIsNotNone(instance.blob)
        self.assertIn("path", instance.blob)
        self.assertEqual(instance.blob["path"], name)

        # Create a form to update the instance but leave the blobfield unchanged
        form = BlankBlobForm(instance=instance, data={"blob": {"path": name}, "category": "test"})
        self.assertTrue(form.is_valid())
        form.save()


class TestCallableBlobField(BlobModelFactoryMixin, StorageOperationsMixin, TestCase):
    """Test the callbacks functionality.
    Use normal test case, allowing on_commit hook to execute.
    """

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_on_change_callback_execution(self):
        with patch("django.db.transaction.on_commit") as mock_on_commit:
            with transaction.atomic():
                # Create a blank instance
                form = CallbackBlobForm(data={"blob": {}})
                instance = form.save()
                instance.refresh_from_db()

                self.assertEqual(mock_on_commit.call_count, 1)

                # Editing while keeping field blank should not execute the callback
                form = CallbackBlobForm(instance=instance, data={"blob": {}})
                instance = form.save()
                self.assertEqual(mock_on_commit.call_count, 1)

                # Editing from blank to valid should execute the callback
                name = self._prefix_blob_name("test_on_change_callback_execution.txt")
                tmp_blob = self._create_temporary_blob(self.bucket)
                form = CallbackBlobForm(
                    instance=instance,
                    data={"blob": {"_tmp_path": tmp_blob.name, "name": name}},
                )
                instance = form.save()
                self.assertEqual(mock_on_commit.call_count, 2)

                # Editing with unchanged valid entry should not trigger a callback
                form = CallbackBlobForm(instance=instance, data={"blob": {"path": name}, "category": "test"})
                form.save()
                self.assertEqual(mock_on_commit.call_count, 2)

                # Editing to a new valid entry should trigger callback
                name = self._prefix_blob_name("test_on_change_callback_execution_2.txt")
                tmp_blob = self._create_temporary_blob(self.bucket)
                form = CallbackBlobForm(
                    instance=instance,
                    data={"blob": {"_tmp_path": tmp_blob.name, "name": name}},
                )
                instance = form.save()
                self.assertEqual(mock_on_commit.call_count, 3)

                # Editing valid to blank should trigger callback
                form = CallbackBlobForm(
                    instance=instance,
                    data={"blob": {}},
                )
                instance = form.save()
                self.assertEqual(mock_on_commit.call_count, 4)

    @override_settings(GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK=get_destination_path_for_test)
    def test_update_attributes_callback_execution(self):
        with self.captureOnCommitCallbacks(execute=True) as callbacks:
            tmp_blob = self._create_temporary_blob(self.bucket)
            name = self._prefix_blob_name("test_attribute_error.txt")

            obj = ExampleUpdateAttributesBlobFieldModel.objects.create(
                category="test", blob={"_tmp_path": tmp_blob.name, "name": name}
            )

            # Callbacks are executed now at the end of this context manager

        obj.refresh_from_db()
        self.assertEqual(len(callbacks), 1)
        blob = get_blob(obj, "blob")
        self.assertEqual(blob.metadata["category"], "test")
        self.assertEqual(blob.content_type, "image/png")


class TestModelBlobField(BlobModelFactoryMixin, StorageOperationsMixin, TestCase):
    """Extra tests for corner cases and model behaviour"""

    def test_no_corruption_with_multiple_models(self):
        """Tests for the presence of a nasty corner case where
        multiple models loaded into memory at once would conflict in
        their behaviour when determining the existing_path
        """

        blob_name_1 = self._prefix_blob_name("no_corruption_1.txt")
        blob_name_2 = self._prefix_blob_name("no_corruption_2.txt")
        # self._create_test_blob(self.bucket, blob_name_1, "")
        # self._create_test_blob(self.bucket, blob_name_2, "")

        # Create two instances
        with override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True):
            obj_1 = ExampleBlobFieldModel.objects.create(blob={"path": blob_name_1})
            obj_1.save()
            # In the captured behaviour, the cache of the second instance overwrite the first
            obj_2 = ExampleBlobFieldModel.objects.create(blob={"path": blob_name_2})
            obj_2.save()

        # Edit the first, without changing the KMZ field (and outside the settings override)
        new_obj_1 = ExampleBlobFieldModel.objects.get(id=obj_1.id)
        new_obj_1.category = "test-change"

        # Note this is not a blankable model, and the observed flaw in flushing
        # the cache in this case manifests in passing a None value instead of
        # the correct existing path being passed through. This save() thus
        # resulted in an IntegrityError.
        new_obj_1.save()
        self.assertEqual(new_obj_1.blob["path"], blob_name_1)



# Disabled for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

import json
from unittest.mock import patch

from django.test import TestCase, override_settings
from django.urls import reverse


def raise_error(*args, **kwargs):
    """Mock to simulate an error during signal handling"""
    raise Exception("An error of some unknown kind, which may happen during signal handling")


class GCloudEventTests(TestCase):
    @patch("django_gcp.events.signals.event_received.send")
    def test_valid_event_is_signalled(self, mock):
        """Ensure a signal is dispatched with the event details"""

        payload = {"the-event": "payload"}
        response = self.client.post(
            reverse("gcp-events", args=["the-event-kind", "the-event-reference"]),
            data=json.dumps(payload),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 201)
        self.assertTrue(mock.called)
        self.assertEqual(mock.call_count, 1)
        self.assertIn("sender", mock.call_args.kwargs)
        self.assertIn("event_kind", mock.call_args.kwargs)
        self.assertIn("event_reference", mock.call_args.kwargs)
        self.assertIn("event_payload", mock.call_args.kwargs)

    def test_post_with_no_reference(self):
        """Ensure that posts with no reference will not be matched"""
        endpoint = reverse("gcp-events", args=["the-event-kind", "anything"]).replace("anything", "")
        response = self.client.post(
            endpoint,
            data="{}",
            content_type="application/json",
        )
        # A catch-all redirect to /admin/ URL comes after the installed urlpatterns, redirecting the user
        # to the admin (for convenience using the test server)
        self.assertEqual(response.status_code, 302)

    def test_non_post_http_methods_disallowed(self):
        """Ensure that methods other than POST are not allowed"""
        endpoint = reverse("gcp-events", args=["the-event-kind", "the-event-reference"])
        response = self.client.get(
            endpoint,
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 405)

    @patch("django_gcp.events.signals.event_received.send", new=raise_error)
    def test_handling_errors_are_returned_unhandleable(self):
        """Ensure that 400 errors are returned if the payload causes an internal error.
        This might not be obvious (since it masks genuine 500 errors), but since the
        payload can be completely arbitrary"""

        response = self.client.post(
            reverse("gcp-events", args=["the-event-kind", "the-event-reference"]),
            data="{}",
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

    @patch("django_gcp.events.signals.event_received.send", new=raise_error)
    def test_handling_errors_are_raised_in_debug_mode(self):
        """Ensure that posting an event when settings.DEBUG=True will raise"""

        with override_settings(DEBUG=True):
            with self.assertRaises(Exception):
                self.client.post(
                    reverse("gcp-events", args=["the-event-kind", "the-event-reference"]),
                    data="{}",
                    content_type="application/json",
                )



# Disables for testing:
# pylint: disable=missing-docstring

import json

from django.db import connection
from django_test_migrations.contrib.unittest_case import MigratorTestCase


class FileToBlobMigrationTestCase(MigratorTestCase):
    """This class is used to test a migration that alters a FileField to a BlobField"""

    migrate_from = ("example", "0001_initial")
    migrate_to = ("example", "0004_rename_blob_temp_exampleblobfieldmodel_blob")

    def prepare(self):
        """Prepare original data at migration 0001, containing the
        string filename for a populated FileField.

        Note that these need to be done with custom SQL entries,
        because models may no longer exist as the repo evolves.

        These statements were generated with the example model and code
        at migration 0001 then using DJANGO DEBUG TOOLBAR with the
        intercept_redirects field set to True, to record the POST
        request for making a model.

        You can also do it programatically
        ```py
            from django.db import connection
            print(connection.queries)
        ```
        shows the sql you executed within the shell context, which gets replicated here
        """

        with connection.cursor() as cursor:
            cursor.execute(
                'INSERT INTO "example_exampleblobfieldmodel" ("blob") VALUES (\'test/file_to_blob_migration.txt\') RETURNING "example_exampleblobfieldmodel"."id"'
            )

    def test_migration_from_file_field(self):
        """Ensure that the string left over"""

        # Ensure an example entry was created
        ExampleBlobFieldModel = self.new_state.apps.get_model("example", "ExampleBlobFieldModel")
        self.assertEqual(ExampleBlobFieldModel.objects.count(), 1)

        # The new model should contain a json string
        row = ExampleBlobFieldModel.objects.first()
        with connection.cursor() as cursor:
            cursor.execute(f'SELECT "blob" FROM "example_exampleblobfieldmodel" WHERE "id" = {row.id}')
            # The first (only) row should contain a decodable json string
            result = json.loads(cursor.fetchone()[0])

        self.assertEqual(result["path"], "test/file_to_blob_migration.txt")



"""GCP Cloud Workflows integration for Django."""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import google.auth
from google.cloud.workflows import executions_v1
from google.cloud.workflows.executions_v1.types import Execution

from django_gcp.tasks.serializers import serialize

from .exceptions import (
    InvalidWorkflowArgumentsError,
    WorkflowConfigurationError,
    WorkflowExecutionError,
    WorkflowNotFoundError,
)


@dataclass
class WorkflowExecution:
    """Wrapper for GCP Workflow execution response.

    Attributes:
        id: The execution ID (last part of the execution name)
        name: Full execution resource name
        workflow: Workflow resource name
        state: Execution state (ACTIVE, SUCCEEDED, FAILED, CANCELLED)
        start_time: When execution started
        end_time: When execution ended (if completed)
        argument: The JSON argument passed to the workflow
        result: The workflow result (if completed successfully)
    """

    id: str
    name: str
    workflow: str
    state: str
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    argument: Optional[str] = None
    result: Optional[str] = None

    @classmethod
    def from_execution(cls, execution: Execution) -> "WorkflowExecution":
        """Create WorkflowExecution from GCP Execution proto."""
        # Extract execution ID from name (format: projects/.../locations/.../workflows/.../executions/{id})
        execution_id = execution.name.split("/")[-1]

        return cls(
            id=execution_id,
            name=execution.name,
            workflow=execution.workflow_revision_id or "",
            state=execution.state.name,
            start_time=execution.start_time,
            end_time=execution.end_time if execution.end_time else None,
            argument=execution.argument if execution.argument else None,
            result=execution.result if execution.result else None,
        )


class Workflow:
    """Base class for triggering GCP Cloud Workflows.

    Subclasses must define:
        - workflow_name: The name of the workflow deployed in GCP
        - location: The GCP region where the workflow is deployed

    Example:
        class ProcessUserData(Workflow):
            workflow_name = "process-user-data"
            location = "europe-west1"

        # Invoke the workflow
        execution = ProcessUserData().invoke(user_id=123, action="process")

        # Store execution ID for tracking
        execution_id = execution.id

        # Later: check status
        status = ProcessUserData().get_execution_status(execution_id)

        # Get console URL
        url = ProcessUserData().get_console_url(execution_id)
    """

    workflow_name: str = None
    location: str = None

    def __init__(self):
        """Initialize the workflow client and get project information from credentials."""
        if not self.workflow_name:
            raise WorkflowConfigurationError(f"{self.__class__.__name__} must define 'workflow_name' class attribute")
        if not self.location:
            raise WorkflowConfigurationError(f"{self.__class__.__name__} must define 'location' class attribute")

        # Get default credentials and project ID
        self._credentials, self._project_id = google.auth.default()

        if not self._project_id:
            raise WorkflowConfigurationError(
                "Could not determine project ID from credentials. Ensure GOOGLE_APPLICATION_CREDENTIALS is set."
            )

        # Initialize the Executions client
        self._client = executions_v1.ExecutionsClient(credentials=self._credentials)

    @property
    def project_id(self) -> str:
        """Get the GCP project ID."""
        return self._project_id

    @property
    def workflow_path(self) -> str:
        """Get the full workflow resource path."""
        return f"projects/{self.project_id}/locations/{self.location}/workflows/{self.workflow_name}"

    def invoke(self, **kwargs) -> WorkflowExecution:
        """Invoke the workflow with the given arguments.

        Args:
            **kwargs: Keyword arguments to pass to the workflow as JSON.
                     These will be serialized using Django's JSON encoder.

        Returns:
            WorkflowExecution: Execution object with ID and initial state.

        Raises:
            InvalidWorkflowArgumentsError: If arguments cannot be serialized.
            WorkflowNotFoundError: If the workflow doesn't exist.
            WorkflowExecutionError: If execution creation fails.
        """
        # Serialize arguments to JSON
        try:
            argument_json = serialize(kwargs) if kwargs else None
        except (TypeError, ValueError) as e:
            raise InvalidWorkflowArgumentsError(f"Failed to serialize workflow arguments: {e}") from e

        # Create execution request
        execution = Execution()
        if argument_json:
            execution.argument = argument_json

        # Invoke the workflow
        try:
            response = self._client.create_execution(
                request={
                    "parent": self.workflow_path,
                    "execution": execution,
                }
            )
        except Exception as e:
            error_msg = str(e)
            if "NOT_FOUND" in error_msg or "404" in error_msg:
                raise WorkflowNotFoundError(
                    f"Workflow '{self.workflow_name}' not found in project '{self.project_id}', "
                    f"location '{self.location}'. Ensure the workflow is deployed."
                ) from e
            raise WorkflowExecutionError(f"Failed to invoke workflow '{self.workflow_name}': {error_msg}") from e

        return WorkflowExecution.from_execution(response)

    def get_execution_status(self, execution_id: str) -> WorkflowExecution:
        """Get the current status of a workflow execution.

        Args:
            execution_id: The execution ID returned from invoke().

        Returns:
            WorkflowExecution: Current execution state and details.

        Raises:
            WorkflowExecutionError: If fetching execution status fails.
        """
        execution_name = f"{self.workflow_path}/executions/{execution_id}"

        try:
            execution = self._client.get_execution(name=execution_name)
        except Exception as e:
            raise WorkflowExecutionError(f"Failed to get execution status for '{execution_id}': {e}") from e

        return WorkflowExecution.from_execution(execution)

    def get_console_url(self, execution_id: str) -> str:
        """Get the GCP Console URL for monitoring a workflow execution.

        Args:
            execution_id: The execution ID returned from invoke().

        Returns:
            str: URL to view the execution in GCP Console.
        """
        # Format: https://console.cloud.google.com/workflows/workflow/{location}/{workflow}/execution/{execution_id}?project={project}
        return (
            f"https://console.cloud.google.com/workflows/workflow/"
            f"{self.location}/{self.workflow_name}/execution/{execution_id}"
            f"?project={self.project_id}"
        )



# Disabled for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

import base64
from datetime import datetime, timezone
import json
from unittest.mock import patch
from zoneinfo import ZoneInfo

from django.test import TestCase

from django_gcp.events.utils import decode_pubsub_message, get_event_url, make_pubsub_message

DEFAULT_SUBSCRIPTION = "projects/my-project/subscriptions/my-subscription"


class GCloudEventUtilsTests(TestCase):
    @patch("django_gcp.events.signals.event_received.send")
    def test_get_event_url_with_parameters(self, mock):
        """Ensure that push endpoint URLs can be reversed successfully with parameters that are decoded on receipt"""

        complex_parameter = "://something?> awkward"
        event_url = get_event_url(
            "the-kind", "the-reference", event_parameters={"complex_parameter": complex_parameter}, base_url=""
        )

        response = self.client.post(
            event_url,
            data="{}",
            content_type="application/json",
        )

        self.assertEqual(response.status_code, 201)
        self.assertTrue(mock.called)
        self.assertEqual(mock.call_count, 1)
        self.assertIn("sender", mock.call_args.kwargs)
        self.assertIn("event_kind", mock.call_args.kwargs)
        self.assertIn("event_reference", mock.call_args.kwargs)
        self.assertIn("event_payload", mock.call_args.kwargs)
        self.assertIn("event_parameters", mock.call_args.kwargs)
        self.assertIn("complex_parameter", mock.call_args.kwargs["event_parameters"])
        self.assertEqual(mock.call_args.kwargs["event_parameters"]["complex_parameter"], complex_parameter)

    def test_endpoint_with_no_base_url(self):
        """Ensure that an AttributeError is correctly raised when getting an event url with no settings.BASE_URL"""

        with self.assertRaises(AttributeError):
            get_event_url("the-kind", "the-reference")

    def test_endpoint_with_base_url(self):
        """Ensure that an AttributeError is correctly raised when getting an event url with no settings.BASE_URL"""

        with self.settings(BASE_URL="https://something.com"):
            event_url = get_event_url("the-kind", "the-reference")
            self.assertEqual(event_url, "https://something.com/example-django-gcp/events/the-kind/the-reference")

    def test_make_bare_pubsub_message(self):
        msg = make_pubsub_message({"my": "data"}, DEFAULT_SUBSCRIPTION)
        decoded = json.loads(msg)
        self.assertIn("message", decoded)
        self.assertIn("subscription", decoded)
        self.assertIn("data", decoded["message"])
        self.assertNotIn("publishTime", decoded["message"])
        self.assertNotIn("orderingKey", decoded["message"])
        self.assertNotIn("messageId", decoded["message"])

    def test_make_full_pubsub_message(self):
        dt = datetime(2022, 8, 12, 9, 0, 25, 226743)
        msg = make_pubsub_message(
            {"my": "data"},
            DEFAULT_SUBSCRIPTION,
            attributes={"one": "two"},
            message_id="myid",
            ordering_key="one",
            publish_time=dt,
        )
        decoded = json.loads(msg)
        self.assertIn("message", decoded)
        self.assertIn("subscription", decoded)
        self.assertIn("data", decoded["message"])
        self.assertIn("publishTime", decoded["message"])
        self.assertIn("orderingKey", decoded["message"])
        self.assertIn("messageId", decoded["message"])
        self.assertEqual(decoded["message"]["data"], "eyJteSI6ICJkYXRhIn0=")
        self.assertEqual(decoded["message"]["publishTime"], "2022-08-12T09:00:25.226743000Z")
        self.assertEqual(decoded["message"]["orderingKey"], "one")
        self.assertEqual(decoded["message"]["messageId"], "myid")

    def test_make_pubsub_message_fails_with_invalid_attributes(self):
        with self.assertRaises(ValueError) as e:
            make_pubsub_message(
                {"my": "data"},
                DEFAULT_SUBSCRIPTION,
                attributes={2: "a"},
            )
        self.assertEqual(e.exception.args[0], "All attribute keys must be strings")

        with self.assertRaises(ValueError) as e:
            make_pubsub_message(
                {"my": "data"},
                DEFAULT_SUBSCRIPTION,
                attributes={"a": 2},
            )
        self.assertEqual(e.exception.args[0], "All attribute values must be strings")

    def test_make_pubsub_message_fails_with_invalid_message_id(self):
        with self.assertRaises(ValueError) as e:
            make_pubsub_message(
                {"my": "data"},
                DEFAULT_SUBSCRIPTION,
                message_id=2,
            )
        self.assertEqual(e.exception.args[0], "The message_id, if given, must be a string")

    def test_make_pubsub_message_fails_with_invalid_ordering_key(self):
        with self.assertRaises(ValueError) as e:
            make_pubsub_message(
                {"my": "data"},
                DEFAULT_SUBSCRIPTION,
                ordering_key=2,
            )
        self.assertEqual(e.exception.args[0], "The ordering_key, if given, must be a string")

    def test_make_pubsub_message_fails_with_invalid_subscription(self):
        with self.assertRaises(ValueError) as e:
            make_pubsub_message(
                {"my": "data"},
                {"subscription-is-not": "a-string"},
            )
        self.assertEqual(
            e.exception.args[0],
            "The subscription must be a string, like 'projects/my-project/subscriptions/my-subscription-name'",
        )

    def test_make_pubsub_message_adjusts_to_utc(self):
        """Ensure that non-naive publish_time datetimes are correctly converted"""

        # LA 7 hours behind UTC at this datetime
        dt = datetime(2022, 8, 12, 9, 0, 25, 226743, tzinfo=ZoneInfo("America/Los_Angeles"))
        utc_dt = datetime(2022, 8, 12, 16, 0, 25, 226743, tzinfo=timezone.utc)

        msg = make_pubsub_message(
            {"my": "data"},
            DEFAULT_SUBSCRIPTION,
            attributes={"one": "two"},
            message_id="myid",
            ordering_key="one",
            publish_time=dt,
        )

        decoded = decode_pubsub_message(msg)

        self.assertEqual(decoded["publish_time"], utc_dt)

    def test_decode_pubsub_message(self):
        dt = datetime(2022, 8, 12, 9, 0, 25, 226743, tzinfo=timezone.utc)
        body = make_pubsub_message(
            {"my": "data"},
            DEFAULT_SUBSCRIPTION,
            attributes={"one": "two"},
            message_id="myid",
            ordering_key="one",
            publish_time=dt,
        )

        decoded = decode_pubsub_message(body)
        self.assertIn("data", decoded)
        self.assertIn("my", decoded["data"])
        self.assertIn("publish_time", decoded)
        self.assertEqual(decoded["publish_time"], dt)

    def test_decode_pubsub_message_with_non_decodable_string_data(self):
        """Test that raw strings in pub/sub messages can be decoded.

        A single string is a valid token in JSON, but it has to be encoded with its quotation marks.
        However, a string that's base-64 encoded directly without being jumped to JSON first is still a valid
        PubSub message.

        That is, endpoints could receive data that's the base 64 encoded verison of:
            b'"stuff"' or
            b'stuff'

        So the decoder has to handle them both the same. We test here that it can.
        """

        body = make_pubsub_message("stuff", DEFAULT_SUBSCRIPTION, as_dict=True)
        body["message"]["data"] = base64.b64encode("stuff".encode()).decode()

        decoded = decode_pubsub_message(body)
        self.assertIn("data", decoded)
        self.assertEqual("stuff", decoded["data"])

    def test_decode_invalid_pubsub_message_raises_value_error(self):
        with self.assertRaises(ValueError) as e:
            body = json.dumps({"has_no_message": {}, "or_subscription": {}}).encode("utf-8")
            decode_pubsub_message(body)

        self.assertIn("Failed to decode Pub/Sub message", e.exception.args[0])



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

from django.test import SimpleTestCase, override_settings

from django_gcp.tasks.tasks import apply_resource_affix


class TasksUtilsTest(SimpleTestCase):
    def test_resource_affix_and_delimiter(self):
        with override_settings(GCP_TASKS_RESOURCE_AFFIX="something", GCP_TASKS_DELIMITER="///"):
            prefixed = apply_resource_affix("value", suffix=False)
            suffixed = apply_resource_affix("value", suffix=True)
            self.assertEqual(prefixed, "something///value")
            self.assertEqual(suffixed, "value///something")

        with override_settings(GCP_TASKS_RESOURCE_AFFIX=None, GCP_TASKS_DELIMITER="///"):
            prefixed = apply_resource_affix("value", suffix=False)
            self.assertEqual(prefixed, "value")



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

# Disabled because gcloud api dynamically constructed
# pylint: disable=no-member

import json
from unittest.mock import patch

from django.test import SimpleTestCase, override_settings
from django.urls import reverse
from google.api_core.exceptions import AlreadyExists

from django_gcp.events.utils import make_pubsub_message
from django_gcp.exceptions import DuplicateTaskError, IncompatibleSettingsError, IncorrectTaskUsageError
from django_gcp.tasks import OnDemandTask
from django_gcp.tasks._pilot.mocker import patch_auth
from tests.server.example.tasks import (
    DeduplicatedOnDemandTask,
    FailingOnDemandTask,
    MyOnDemandTask,
    MyPeriodicTask,
    MySubscriberTask,
)

from .test_events_utils import DEFAULT_SUBSCRIPTION


class TasksEnqueueingTest(SimpleTestCase):
    def test_instantiate_task_directly(self):
        with self.assertRaises(IncorrectTaskUsageError):
            OnDemandTask()

    def test_enqueue_duplicatable_on_demand_task(self):
        with patch_auth():
            with patch("django_gcp.tasks._pilot.tasks.CloudTasks.push"):
                MyOnDemandTask().enqueue(a="1")

    def test_enqueue_deduplicated_task_raises_exception_on_duplicate(self):
        """Ensures that a unique task cannot be enqueued"""

        with patch_auth():
            with patch("django_gcp.tasks._pilot.tasks.CloudTasks.push") as patched_push:
                patched_push.side_effect = AlreadyExists("409 Requested entity already exists")

                with self.assertRaises(DuplicateTaskError):
                    DeduplicatedOnDemandTask().enqueue(a="1")

    def test_periodic_task_run_method_called_with_data(self):
        url = reverse("gcp-tasks", args=["MyPeriodicTask"])
        data = {"a": 1}
        patch_response = {"b": 2}
        with patch("tests.server.example.tasks.MyPeriodicTask.run", return_value=patch_response) as patched_run:
            response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(200, response.status_code)
        self.assertEqual({"result": patch_response}, response.json())
        patched_run.assert_called_once_with(**data)

    def test_invalid_task_name(self):
        url = reverse("gcp-tasks", args=["NotAValidTaskName"])
        data = {"a": 1}
        response = self.client.post(path=url, data=json.dumps(data), content_type="application/json")

        self.assertEqual(404, response.status_code)
        self.assertIn("error", response.json())
        self.assertEqual("Task NotAValidTaskName not found", response.json()["error"])

    def test_subscriber_task_run_method_called_with_data(self):
        url = reverse("gcp-subscriber-tasks", args=["MySubscriberTask"])

        msg = make_pubsub_message({"a": 1}, DEFAULT_SUBSCRIPTION)
        with patch("tests.server.example.tasks.MySubscriberTask.run", return_value=None) as patched_run:
            response = self.client.post(path=url, data=msg, content_type="application/json")

        self.assertEqual(200, response.status_code)
        self.assertEqual({"result": None}, response.json())
        patched_run.assert_called_once()

    def test_disable_enqueueing_with_a_setting(self):
        """Assert that no task is enqueued if the GCP_TASKS_DISABLE_EXECUTE is true"""
        with patch_auth():
            with patch("django_gcp.tasks.tasks.run_coroutine") as patched_send:
                MyOnDemandTask().enqueue(a="1")
                self.assertEqual(patched_send.call_count, 1)

                with override_settings(GCP_TASKS_DISABLE_EXECUTE=True):
                    MyOnDemandTask().enqueue(a="1")
                    MyPeriodicTask().enqueue(a="1")
                    MySubscriberTask().enqueue(a="1")
                    FailingOnDemandTask().enqueue(a="1")

                    self.assertEqual(patched_send.call_count, 1)

                    MyOnDemandTask().enqueue_later(a="1", when=10)
                    MyPeriodicTask().enqueue_later(a="1", when=10)
                    MySubscriberTask().enqueue_later(a="1", when=10)
                    FailingOnDemandTask().enqueue_later(a="1", when=10)

                    self.assertEqual(patched_send.call_count, 1)

            with override_settings(GCP_TASKS_DISABLE_EXECUTE=True, GCP_TASKS_EAGER_EXECUTE=True):
                with self.assertRaises(IncompatibleSettingsError):
                    MyOnDemandTask().enqueue(a="1")



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

from io import StringIO
from typing import List
from unittest.mock import Mock, patch

from django.core.management import call_command
from django.test import SimpleTestCase, override_settings

from django_gcp.exceptions import UnknownActionError
from django_gcp.tasks._pilot.mocker import patch_auth


class CommandsTest(SimpleTestCase):
    def patch_schedule(self, **kwargs):
        return patch("django_gcp.tasks._pilot.scheduler.CloudScheduler.put", **kwargs)

    def patch_subscribe(self, **kwargs):
        return patch("django_gcp.tasks._pilot.pubsub.CloudSubscriber.create_subscription", **kwargs)

    def patch_get_scheduled(self, names: List[str] = None, **kwargs):
        jobs = []
        for name in names or []:
            job = Mock()
            job.name = f"/app/jobs/{name}"
            jobs.append(job)
        return patch("django_gcp.tasks._pilot.scheduler.CloudScheduler.list", return_value=jobs, **kwargs)

    def patch_delete_schedule(self):
        return patch("django_gcp.tasks._pilot.scheduler.CloudScheduler.delete")

    def _assert_command(
        self,
        command: str,
        params: List[str] = None,
        expected_schedule_calls: int = 0,
        expected_subscribe_calls: int = 0,
        expected_get_scheduled_calls: int = 0,
        expected_output: str = None,
        scheduled_job_names: List[str] = None,
    ):
        out = StringIO()
        with patch_auth():
            with self.patch_schedule() as schedule:
                with self.patch_subscribe() as subscribe:
                    scheduled_job_names = scheduled_job_names or []
                    with self.patch_get_scheduled(names=scheduled_job_names) as get_scheduled:
                        call_params = params or []
                        call_command(command, *call_params, no_color=True, stdout=out)

        self.assertEqual(expected_schedule_calls, schedule.call_count)
        self.assertEqual(expected_subscribe_calls, subscribe.call_count)
        self.assertEqual(expected_get_scheduled_calls, get_scheduled.call_count)
        if expected_output:
            self.assertEqual(expected_output, out.getvalue())

        return out.getvalue()

    def test_unknown_manager_action(self):
        out = StringIO()
        with self.assertRaises(UnknownActionError):
            call_command("task_manager", "someActionThatDoesntExist", no_color=True, stdout=out)

    def test_create_scheduler_jobs(self):
        with override_settings(GCP_TASKS_DOMAIN="https://wherever.com"):
            out = self._assert_command(
                command="task_manager",
                params=["create_scheduler_jobs"],
                expected_schedule_calls=1,
                expected_get_scheduled_calls=0,
            )
        self.assertIn(
            "Successfully created 1 scheduler jobs to domain https://wherever.com",
            out,
        )
        self.assertIn(
            "- [+] django-gcp--myperiodictask",
            out,
        )

    def test_create_scheduler_jobs_with_no_cleanup_affix(self):
        with override_settings(GCP_TASKS_RESOURCE_AFFIX=None):
            with override_settings(GCP_TASKS_DOMAIN="https://wherever.com"):
                self._assert_command(
                    command="task_manager",
                    params=["create_scheduler_jobs", "--cleanup"],
                    expected_schedule_calls=1,
                    expected_get_scheduled_calls=0,
                )
        # TODO ensure that the right warning appears in logs
        # self.assertIn(
        #     "Cleanup of unused Scheduler Jobs was requested but skipped, because no resource affix was given - see the GCP_TASKS_RESOURCE_AFFIX docs for more details.",
        #     out,
        # )

    def test_create_scheduler_jobs_with_cleanup_affix(self):
        with override_settings(GCP_TASKS_RESOURCE_AFFIX="notnone"):
            with override_settings(GCP_TASKS_DOMAIN="https://wherever.com"):
                self._assert_command(
                    command="task_manager",
                    params=["create_scheduler_jobs", "--cleanup"],
                    expected_schedule_calls=1,
                    expected_get_scheduled_calls=1,
                )
        # TODO mock the response from get_expected and assert that the deletion is called
        # It might look a bit like this:
        # from unittest.mock import patch, Mock, PropertyMock
        #     expected_output = (
        #         "Successfully scheduled 3 tasks to domain http://localhost:8080"
        #         "\n- [+] SaySomethingTask"
        #         "\n- [-] potato_task_1"
        #         "\n- [-] potato_task_2\n"
        #     )
        #     names = ["potato_task_1", "potato_task_2"]
        #     app_config = apps.get_app_config("django_cloud_tasks")
        #     with patch.object(app_config, "app_name", new_callable=PropertyMock, return_value="potato"):
        #         with self.patch_get_scheduled(names=names):
        #             with self.patch_delete_schedule() as delete:
        #                 self._assert_command(
        #                     command="schedule_tasks",
        #                     expected_schedule_calls=1,
        #                     expected_output=expected_output,
        #                 )
        #             self.assertEqual(2, delete.call_count)

    def test_create_pubsub_subscriptions(self):
        with override_settings(GCP_TASKS_DOMAIN="https://wherever.com"):
            out = self._assert_command(
                command="task_manager",
                params=["create_pubsub_subscriptions"],
                expected_subscribe_calls=1,
                expected_schedule_calls=0,
                expected_get_scheduled_calls=0,
            )
        self.assertEqual(
            "Successfully created 1 pubsub subscriptions to domain https://wherever.com\n- [+] MySubscriberTask\n",
            out,
        )



"""Custom exceptions for GCP Cloud Workflows."""


class WorkflowError(Exception):
    """Base exception for workflow-related errors."""

    pass


class WorkflowNotFoundError(WorkflowError):
    """Raised when a workflow cannot be found in GCP."""

    pass


class WorkflowExecutionError(WorkflowError):
    """Raised when workflow execution fails."""

    pass


class InvalidWorkflowArgumentsError(WorkflowError):
    """Raised when workflow arguments are invalid or cannot be serialized."""

    pass


class WorkflowConfigurationError(WorkflowError):
    """Raised when workflow configuration is invalid or incomplete."""

    pass



# Disables for testing:
# pylint: disable=missing-docstring
# pylint: disable=protected-access

from datetime import datetime
from unittest.mock import MagicMock, patch

from django.test import SimpleTestCase

from django_gcp.workflows import (
    InvalidWorkflowArgumentsError,
    Workflow,
    WorkflowConfigurationError,
    WorkflowExecution,
    WorkflowExecutionError,
    WorkflowNotFoundError,
)


class SampleWorkflow(Workflow):
    """Sample workflow class for unit tests."""

    workflow_name = "test-workflow"
    location = "europe-west1"


class WorkflowExecutionTest(SimpleTestCase):
    """Tests for WorkflowExecution dataclass."""

    def test_from_execution_basic(self):
        """Test creating WorkflowExecution from GCP Execution proto."""
        mock_execution = MagicMock()
        mock_execution.name = "projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-123"
        mock_execution.workflow_revision_id = "test-workflow-rev-1"
        mock_execution.state.name = "ACTIVE"
        mock_execution.start_time = datetime(2025, 1, 1, 12, 0, 0)
        mock_execution.end_time = None
        mock_execution.argument = '{"key": "value"}'
        mock_execution.result = None

        workflow_execution = WorkflowExecution.from_execution(mock_execution)

        self.assertEqual(workflow_execution.id, "exec-123")
        self.assertEqual(workflow_execution.name, mock_execution.name)
        self.assertEqual(workflow_execution.workflow, "test-workflow-rev-1")
        self.assertEqual(workflow_execution.state, "ACTIVE")
        self.assertEqual(workflow_execution.start_time, datetime(2025, 1, 1, 12, 0, 0))
        self.assertIsNone(workflow_execution.end_time)
        self.assertEqual(workflow_execution.argument, '{"key": "value"}')
        self.assertIsNone(workflow_execution.result)

    def test_from_execution_completed(self):
        """Test creating WorkflowExecution from completed execution."""
        mock_execution = MagicMock()
        mock_execution.name = "projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-456"
        mock_execution.workflow_revision_id = "test-workflow-rev-1"
        mock_execution.state.name = "SUCCEEDED"
        mock_execution.start_time = datetime(2025, 1, 1, 12, 0, 0)
        mock_execution.end_time = datetime(2025, 1, 1, 12, 5, 0)
        mock_execution.argument = '{"input": "data"}'
        mock_execution.result = '{"output": "result"}'

        workflow_execution = WorkflowExecution.from_execution(mock_execution)

        self.assertEqual(workflow_execution.id, "exec-456")
        self.assertEqual(workflow_execution.state, "SUCCEEDED")
        self.assertEqual(workflow_execution.end_time, datetime(2025, 1, 1, 12, 5, 0))
        self.assertEqual(workflow_execution.result, '{"output": "result"}')


class WorkflowConfigurationTest(SimpleTestCase):
    """Tests for Workflow configuration and initialization."""

    @patch("google.auth.default")
    def test_workflow_requires_workflow_name(self, mock_auth):
        """Test that Workflow class requires workflow_name attribute."""
        mock_auth.return_value = (MagicMock(), "test-project")

        class InvalidWorkflow(Workflow):
            location = "europe-west1"

        with self.assertRaises(WorkflowConfigurationError) as context:
            InvalidWorkflow()

        self.assertIn("workflow_name", str(context.exception))

    @patch("google.auth.default")
    def test_workflow_requires_location(self, mock_auth):
        """Test that Workflow class requires location attribute."""
        mock_auth.return_value = (MagicMock(), "test-project")

        class InvalidWorkflow(Workflow):
            workflow_name = "test-workflow"

        with self.assertRaises(WorkflowConfigurationError) as context:
            InvalidWorkflow()

        self.assertIn("location", str(context.exception))

    @patch("google.auth.default")
    def test_workflow_requires_project_id_from_credentials(self, mock_auth):
        """Test that Workflow raises error if project_id cannot be determined."""
        mock_auth.return_value = (MagicMock(), None)

        with self.assertRaises(WorkflowConfigurationError) as context:
            SampleWorkflow()

        self.assertIn("project ID", str(context.exception))

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_workflow_initializes_with_credentials(self, mock_auth, mock_client_class):
        """Test that Workflow initializes ExecutionsClient with credentials."""
        mock_credentials = MagicMock()
        mock_auth.return_value = (mock_credentials, "test-project")

        workflow = SampleWorkflow()

        self.assertEqual(workflow.project_id, "test-project")
        self.assertEqual(workflow.workflow_path, "projects/test-project/locations/europe-west1/workflows/test-workflow")
        mock_client_class.assert_called_once_with(credentials=mock_credentials)


class WorkflowInvokeTest(SimpleTestCase):
    """Tests for Workflow.invoke() method."""

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_invoke_workflow_success(self, mock_auth, mock_client_class):
        """Test successfully invoking a workflow."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        # Mock the execution response
        mock_response = MagicMock()
        mock_response.name = "projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-123"
        mock_response.workflow_revision_id = "test-workflow-rev-1"
        mock_response.state.name = "ACTIVE"
        mock_response.start_time = datetime(2025, 1, 1, 12, 0, 0)
        mock_response.end_time = None
        mock_response.argument = '{"user_id": 123, "action": "process"}'
        mock_response.result = None
        mock_client.create_execution.return_value = mock_response

        workflow = SampleWorkflow()
        result = workflow.invoke(user_id=123, action="process")

        self.assertIsInstance(result, WorkflowExecution)
        self.assertEqual(result.id, "exec-123")
        self.assertEqual(result.state, "ACTIVE")

        # Verify create_execution was called correctly
        mock_client.create_execution.assert_called_once()
        call_kwargs = mock_client.create_execution.call_args.kwargs
        self.assertEqual(
            call_kwargs["request"]["parent"], "projects/test-project/locations/europe-west1/workflows/test-workflow"
        )
        # Check that argument was set (exact JSON format may vary)
        self.assertIn("user_id", call_kwargs["request"]["execution"].argument)

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_invoke_workflow_without_arguments(self, mock_auth, mock_client_class):
        """Test invoking a workflow without arguments."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_response = MagicMock()
        mock_response.name = "projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-456"
        mock_response.workflow_revision_id = ""
        mock_response.state.name = "ACTIVE"
        mock_response.start_time = datetime(2025, 1, 1, 12, 0, 0)
        mock_response.end_time = None
        mock_response.argument = None
        mock_response.result = None
        mock_client.create_execution.return_value = mock_response

        workflow = SampleWorkflow()
        result = workflow.invoke()

        self.assertEqual(result.id, "exec-456")
        mock_client.create_execution.assert_called_once()

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_invoke_workflow_not_found(self, mock_auth, mock_client_class):
        """Test invoking a workflow that doesn't exist."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_client.create_execution.side_effect = Exception("404 NOT_FOUND: Workflow not found")

        workflow = SampleWorkflow()

        with self.assertRaises(WorkflowNotFoundError) as context:
            workflow.invoke(user_id=123)

        self.assertIn("test-workflow", str(context.exception))
        self.assertIn("test-project", str(context.exception))
        self.assertIn("europe-west1", str(context.exception))

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_invoke_workflow_execution_error(self, mock_auth, mock_client_class):
        """Test handling generic execution errors."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_client.create_execution.side_effect = Exception("Some other error")

        workflow = SampleWorkflow()

        with self.assertRaises(WorkflowExecutionError) as context:
            workflow.invoke(user_id=123)

        self.assertIn("test-workflow", str(context.exception))

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_invoke_workflow_with_non_serializable_arguments(self, mock_auth, mock_client_class):
        """Test invoking a workflow with arguments that cannot be serialized to JSON."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        workflow = SampleWorkflow()

        # Pass a non-serializable object (lambda function)
        with self.assertRaises(InvalidWorkflowArgumentsError) as context:
            workflow.invoke(callback=lambda x: x)

        # Verify error message mentions serialization
        self.assertIn("serialize", str(context.exception).lower())

        # Verify the original exception is preserved in the chain
        self.assertIsNotNone(context.exception.__cause__)


class WorkflowStatusTest(SimpleTestCase):
    """Tests for Workflow.get_execution_status() method."""

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_get_execution_status_success(self, mock_auth, mock_client_class):
        """Test getting execution status successfully."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_execution = MagicMock()
        mock_execution.name = "projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-123"
        mock_execution.workflow_revision_id = "test-workflow-rev-1"
        mock_execution.state.name = "SUCCEEDED"
        mock_execution.start_time = datetime(2025, 1, 1, 12, 0, 0)
        mock_execution.end_time = datetime(2025, 1, 1, 12, 5, 0)
        mock_execution.argument = '{"user_id": 123}'
        mock_execution.result = '{"status": "completed"}'
        mock_client.get_execution.return_value = mock_execution

        workflow = SampleWorkflow()
        status = workflow.get_execution_status("exec-123")

        self.assertEqual(status.id, "exec-123")
        self.assertEqual(status.state, "SUCCEEDED")
        self.assertEqual(status.result, '{"status": "completed"}')

        mock_client.get_execution.assert_called_once_with(
            name="projects/test-project/locations/europe-west1/workflows/test-workflow/executions/exec-123"
        )

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_get_execution_status_error(self, mock_auth, mock_client_class):
        """Test handling errors when getting execution status."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client = MagicMock()
        mock_client_class.return_value = mock_client

        mock_client.get_execution.side_effect = Exception("Execution not found")

        workflow = SampleWorkflow()

        with self.assertRaises(WorkflowExecutionError) as context:
            workflow.get_execution_status("exec-invalid")

        self.assertIn("exec-invalid", str(context.exception))


class WorkflowConsoleUrlTest(SimpleTestCase):
    """Tests for Workflow.get_console_url() method."""

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_get_console_url(self, mock_auth, mock_client_class):
        """Test generating GCP Console URL."""
        mock_auth.return_value = (MagicMock(), "test-project")
        mock_client_class.return_value = MagicMock()

        workflow = SampleWorkflow()
        url = workflow.get_console_url("exec-123")

        expected_url = (
            "https://console.cloud.google.com/workflows/workflow/"
            "europe-west1/test-workflow/execution/exec-123"
            "?project=test-project"
        )
        self.assertEqual(url, expected_url)

    @patch("django_gcp.workflows.workflows.executions_v1.ExecutionsClient")
    @patch("google.auth.default")
    def test_get_console_url_different_location(self, mock_auth, mock_client_class):
        """Test console URL generation with different location."""
        mock_auth.return_value = (MagicMock(), "my-project")
        mock_client_class.return_value = MagicMock()

        class UsWorkflow(Workflow):
            workflow_name = "us-workflow"
            location = "us-central1"

        workflow = UsWorkflow()
        url = workflow.get_console_url("exec-456")

        expected_url = (
            "https://console.cloud.google.com/workflows/workflow/"
            "us-central1/us-workflow/execution/exec-456"
            "?project=my-project"
        )
        self.assertEqual(url, expected_url)
